import requests
import re


def get_ip() -> str:
    """Get the IP address of the current machine."""
    try:
        ip = requests.get("https://api.ipify.org").text
    except:
        return ""
    return ip


def get_cloud_region(ip: str, provider: str = "aws") -> str:
    """Get the cloud region which is hosting the current machine
    or closest to the current machine."""
    # todo: implement cloest region
    default_region = {"aws": "us-east-1", "azure": "eastus", "gcp": "us-east1"}
    try:
        if provider == "aws":
            region = requests.get(f"https://ip-ranges.amazonaws.com/ip-ranges.json").json()
            for prefix in region["prefixes"]:
                if re.match(prefix["ip_prefix"], ip):
                    return prefix["region"]
        elif provider == "azure":
            user_agent = {
                "User-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95"
            }
            body = requests.get(f"https://www.microsoft.com/en-us/download/confirmation.aspx?id=56519", headers=user_agent).content
            matches = re.search(b'downloadretry" href="([^"]*)"', body)
            if matches is None:
                return default_region[provider]
            region_url = matches.groups(0)[0]
            region = requests.get(region_url).json()
            for prefix in region["values"]:
                if re.match(prefix["properties"]["addressPrefix"], ip):
                    return prefix["properties"]["region"]
        elif provider == "gcp":
            region = requests.get(f"https://www.gstatic.com/ipranges/cloud.json").json()
            for prefix in region["prefixes"]:
                if re.match(prefix["ipv4Prefix"], ip):
                    return prefix["region"]
    except:
        return default_region[provider]
    return default_region[provider]



class GCPPricing:
    @staticmethod
    def get_transfer_cost(src_key, dst_key, premium_tier=True):
        src_provider, src = src_key.split(":")
        dst_provider, dst = dst_key.split(":")
        assert src_provider == "gcp"
        src_continent, src_region, src_zone = src.split("-")
        if dst_provider == "gcp":
            dst_continent, dst_region, dst_zone = dst.split("-")
            if src_continent == dst_continent and src_region == dst_region and src_zone == dst_zone:
                return 0.0
            elif src_continent == dst_continent and src_region == dst_region:
                return 0.01
            elif src_continent == dst_continent:
                if src_continent == "northamerica" or src_continent == "us":
                    return 0.01
                elif src_continent == "europe":
                    return 0.02
                elif src_continent == "asia":
                    return 0.05
                elif src_continent == "southamerica":
                    return 0.08
                elif src_continent == "australia":
                    return 0.08
                else:
                    raise Exception(f"Unknown continent {src_continent}")
            elif src.startswith("asia-southeast2") or src_continent == "australia":
                return 0.15
            elif dst.startswith("asia-southeast2") or dst_continent == "australia":
                return 0.15
            else:
                return 0.08
        elif dst_provider in ["aws", "azure", "cloudflare"] and premium_tier:
            is_dst_australia = (dst == "ap-southeast-2") if dst_provider == "aws" else (dst.startswith("australia"))
            # singapore or tokyo or osaka
            if src_continent == "asia" and (src_region == "southeast2" or src_region == "northeast1" or src_region == "northeast2"):
                return 0.19 if is_dst_australia else 0.14
            # jakarta
            elif (src_continent == "asia" and src_region == "southeast1") or (src_continent == "australia"):
                return 0.19
            # seoul
            elif src_continent == "asia" and src_region == "northeast3":
                return 0.19 if is_dst_australia else 0.147
            else:
                return 0.19 if is_dst_australia else 0.12
        elif dst_provider in ["aws", "azure", "cloudflare"] and not premium_tier:
            if src_continent == "us" or src_continent == "europe" or src_continent == "northamerica":
                return 0.085
            elif src_continent == "southamerica" or src_continent == "australia":
                return 0.12
            elif src_continent == "asia":
                return 0.11
            else:
                raise ValueError("Unknown src_continent: {}".format(src_continent))



import functools
import importlib


def inject(*modules, pip_extra=None):
    """
    Decorator for dependency injection
    @inject("google.auth")
    def example_fn(auth):
        auth.discovery.build()
    """

    def wrapper(fn):
        @functools.wraps(fn)
        def wrapped(*args, **kwargs):
            modules_imported = []
            for module in modules:
                err_msg = f"Cannot import {module}."
                if pip_extra:
                    err_msg += f" Install skyplane with {pip_extra} support: `pip install 'skyplane[{pip_extra.lower()}]'`"
                try:
                    modules_imported.append(importlib.import_module(module))
                except ImportError as e:
                    module_split = module.split(".")
                    if len(module_split) > 1:  # try the "from x import y" syntax
                        try:
                            module_imported = importlib.import_module(".".join(module_split[:-1]))
                            modules_imported.append(getattr(module_imported, module_split[-1]))
                        except (ImportError, AttributeError):
                            raise ImportError(err_msg) from e
                    else:
                        raise ImportError(err_msg) from e
            return fn(*modules_imported, *args, **kwargs)

        return wrapped

    return wrapper



"""
Cloud convenience interface
"""

import json
import uuid
import subprocess
import time
from collections import defaultdict

import typer
from rich import print as rprint
from typing import List

from skyplane import compute
from skyplane.config_paths import cloud_config
from skyplane.obj_store.s3_interface import S3Interface
from skyplane.utils import logger
from skyplane.utils.definitions import GB
from skyplane.utils.fn import do_parallel
from skyplane.utils.timer import Timer

app = typer.Typer(name="skyplane-cloud")

# Common utils


def check_assert(condition, msg, debug_msg=None):
    if not condition:
        rprint(f"[red][bold]:x: Check failed:[/bold] {msg}[/red]")
        if debug_msg:
            rprint(f"[red]{debug_msg}[/red]")
        raise typer.Exit(1)
    else:
        rprint(f"[green][bold]:heavy_check_mark: Check passed:[/bold] {msg}[/green]")


def run_cmd(cmd, debug: bool):
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    if debug:
        rprint(f"[bright_black] $ {' '.join(cmd)} [bold]=> {process.returncode}[/bold][/bright_black]")
    return process.returncode, stdout.decode("utf-8"), stderr.decode("utf-8")


# AWS CLI tools


@app.command()
def aws_vcpu_limits(quota_code="L-1216C47A"):
    """List the vCPU limits for each region."""
    aws_auth = compute.AWSAuthentication()

    def get_service_quota(region):
        service_quotas = aws_auth.get_boto3_client("service-quotas", region)
        try:
            response = service_quotas.get_service_quota(ServiceCode="ec2", QuotaCode=quota_code)
        except Exception as e:
            logger.exception(e, print_traceback=False)
            logger.error(f"Failed to get service quota for {quota_code} in {region}")
            return -1
        return response["Quota"]["Value"]

    quotas = do_parallel(get_service_quota, compute.AWSCloudProvider.region_list())
    for region, quota in quotas:
        typer.secho(f"{region}: {int(quota)}", fg="green")


@app.command()
def aws_datasync(src_bucket: str, dst_bucket: str, path: str):
    aws_auth = compute.AWSAuthentication()
    src_region = S3Interface(src_bucket).aws_region
    dst_region = S3Interface(dst_bucket).aws_region

    iam_client = aws_auth.get_boto3_client("iam", "us-east-1")
    try:
        response = iam_client.get_role(RoleName="datasync-role")
        typer.secho("IAM role exists datasync-role", fg="green")
    except iam_client.exceptions.NoSuchEntityException:
        typer.secho("Creating datasync-role", fg="green")
        policy = {
            "Version": "2012-10-17",
            "Statement": [{"Effect": "Allow", "Principal": {"Service": "datasync.amazonaws.com"}, "Action": "sts:AssumeRole"}],
        }
        response = iam_client.create_role(RoleName="datasync-role", AssumeRolePolicyDocument=json.dumps(policy))
    iam_client.attach_role_policy(RoleName="datasync-role", PolicyArn="arn:aws:iam::aws:policy/AWSDataSyncFullAccess")
    # attach s3:ListBucket to datasync-role
    iam_client.attach_role_policy(RoleName="datasync-role", PolicyArn="arn:aws:iam::aws:policy/AmazonS3FullAccess")

    iam_arn = response["Role"]["Arn"]
    typer.secho(f"IAM role ARN: {iam_arn}", fg="green")

    # wait for role to be ready
    typer.secho("Waiting for IAM role to be ready", fg="green")
    iam_client.get_waiter("role_exists").wait(RoleName="datasync-role")

    ds_client_src = aws_auth.get_boto3_client("datasync", src_region)
    src_response = ds_client_src.create_location_s3(
        S3BucketArn=f"arn:aws:s3:::{src_bucket}", Subdirectory=path, S3Config={"BucketAccessRoleArn": iam_arn}
    )
    src_s3_arn = src_response["LocationArn"]
    ds_client_dst = aws_auth.get_boto3_client("datasync", dst_region)
    dst_response = ds_client_dst.create_location_s3(
        S3BucketArn=f"arn:aws:s3:::{dst_bucket}", Subdirectory=path, S3Config={"BucketAccessRoleArn": iam_arn}
    )
    dst_s3_arn = dst_response["LocationArn"]

    try:
        create_task_response = ds_client_dst.create_task(
            SourceLocationArn=src_s3_arn,
            DestinationLocationArn=dst_s3_arn,
            Name=f"{src_bucket}-{dst_bucket}-{path}",
            Options={"BytesPerSecond": -1, "OverwriteMode": "ALWAYS", "TransferMode": "ALL", "VerifyMode": "NONE"},
        )
        task_arn = create_task_response["TaskArn"]
    except ds_client_dst.exceptions.InvalidRequestException:
        typer.secho(f"Region not supported: {src_region} to {dst_region}", fg="red", err=True)
        raise typer.Abort()

    with Timer() as t:
        exec_response = ds_client_dst.start_task_execution(TaskArn=task_arn)
        task_execution_arn = exec_response["TaskExecutionArn"]

        def exit():
            task_execution_response = ds_client_dst.describe_task_execution(TaskExecutionArn=task_execution_arn)
            if task_execution_response["Status"] != "SUCCESS":
                ds_client_dst.cancel_task_execution(TaskExecutionArn=task_execution_arn)
                typer.secho("Cancelling task", fg="red", err=True)

        last_status = None
        try:
            while last_status != "SUCCESS":
                task_execution_response = ds_client_dst.describe_task_execution(TaskExecutionArn=task_execution_arn)
                last_status = task_execution_response["Status"]
                metadata_fields = ["EstimatedBytesToTransfer", "BytesWritten", "Result"]
                metadata = {k: v for k, v in task_execution_response.items() if k in metadata_fields}
                typer.secho(f"{int(t.elapsed)}s\tStatus: {last_status}, {metadata}", fg="green")
                time.sleep(5)
                if (int(t.elapsed) > 300) and last_status == "LAUNCHING":
                    typer.secho(
                        "The process might have errored out. Try deleting the objects if they exist already and restart the transfer.",
                        fg="red",
                        err=True,
                    )
        except KeyboardInterrupt:
            if last_status != "SUCCESS":
                exit()

    task_execution_response = ds_client_dst.describe_task_execution(TaskExecutionArn=task_execution_arn)
    transfer_size_gb = task_execution_response["BytesTransferred"] / GB
    transfer_duration_s = task_execution_response["Result"]["TransferDuration"] / 1000
    gbps = transfer_size_gb * 8 / transfer_duration_s
    typer.secho(f"DataSync response: {task_execution_response}", fg="green")
    typer.secho(
        json.dumps(dict(transfer_size_gb=transfer_size_gb, transfer_duration_s=transfer_duration_s, gbps=gbps, total_runtime_s=t.elapsed)),
        fg="white",
    )


@app.command()
def azure_check(
    account: str = typer.Argument(..., help="Storage account name"),
    container: str = typer.Argument(..., help="Container name"),
    debug: bool = typer.Option(False, "--debug", help="Print debug info"),
):
    hline = "=" * 80
    rprint(f"{hline}\n[bold]Checking Skyplane configuration...[/bold]\n{hline}")
    if debug:
        rprint(f"[bright_black]Skyplane config: {cloud_config}[/bright_black]")
    check_assert(cloud_config.azure_enabled, "Azure enabled in config")
    check_assert(cloud_config.azure_principal_id, "Azure principal ID set in config")
    check_assert(cloud_config.azure_subscription_id, "Azure subscription ID set in config")
    check_assert(cloud_config.azure_client_id, "Azure client ID set in config")

    rprint(f"\n{hline}\n[bold]Checking Azure CLI...[/bold]\n{hline}")
    # check that azure cli is installed
    retcode, stdout, stderr = run_cmd(["az", "--version"], debug=debug)
    check_assert(retcode == 0, "Azure CLI installed", debug_msg=stderr)
    if debug:
        rprint(f"[bright_black]Azure CLI version: {stdout.strip()}[/bright_black]")

    # check that azure cli is logged in
    retcode, stdout, stderr = run_cmd(["az", "account", "show"], debug=debug)
    check_assert(retcode == 0, "Azure CLI logged in", debug_msg=stderr)
    if debug:
        rprint(f"[bright_black]Azure CLI logged in as {stdout}[/bright_black]")

    # check that azure cli has correct subscription
    retcode, stdout, stderr = run_cmd(["az", "account", "show", "--query", "id"], debug=debug)
    if debug:
        rprint(f"[bright_black]Azure CLI subscription: {stdout.strip()}[/bright_black]")
    if debug:
        rprint(f"[bright_black]Skyplane subscription: {cloud_config.azure_subscription_id}[/bright_black]")
    check_assert(retcode == 0, "Azure CLI has subscription set", debug_msg=stderr)
    check_assert(stdout.replace('"', "").strip() == cloud_config.azure_subscription_id, "Azure CLI has correct subscription set")

    # check Azure UMIs
    rprint(f"\n{hline}\n[bold]Checking Azure UMIs...[/bold]\n{hline}")
    # list all UMIs with CLI
    cli = f"az identity list --resource-group skyplane"
    retcode, stdout, stderr = run_cmd(cli.split(), debug=debug)
    check_assert(retcode == 0, "Azure CLI UMIs listed", debug_msg=stderr)
    if debug:
        rprint(f"[bright_black]Azure CLI UMIs: {stdout}[/bright_black]")
    # check Skyplane UMI is in list
    parsed = json.loads(stdout)
    matched_umi_idx = [i for i, umi in enumerate(parsed) if umi["name"] == "skyplane_umi"]
    check_assert(len(matched_umi_idx) == 1, f"Skyplane UMI exists")
    umi = parsed[matched_umi_idx[0]]
    if debug:
        rprint(f"[bright_black]Skyplane UMI: {umi}[/bright_black]")
    check_assert(umi["clientId"] == cloud_config.azure_client_id, "Skyplane UMI has correct client ID")
    check_assert(umi["principalId"] == cloud_config.azure_principal_id, "Skyplane UMI has correct principal ID")
    if debug:
        rprint(f"[bright_black]Skyplane UMI tenant ID: {umi['tenantId']}[/bright_black]")

    # check that Python SDK
    auth = compute.AzureAuthentication(cloud_config)
    rprint(f"\n{hline}\n[bold]Checking Azure Python SDK...[/bold]\n{hline}")
    cred = auth.credential
    if debug:
        rprint(f"[bright_black]Azure Python SDK credential: {cred}[/bright_black]")
    check_assert(cred, "Azure Python SDK credential created")
    from azure.identity import DefaultAzureCredential

    check_assert(isinstance(cred, DefaultAzureCredential), "Azure Python SDK credential is DefaultAzureCredential")
    token = auth.get_token("https://storage.azure.com/.default")
    check_assert(token, "Azure Python SDK token created")

    # check that storage management client works
    rprint(f"\n{hline}\n[bold]Checking Azure storage management client...[/bold]\n{hline}")
    storage_client = auth.get_storage_management_client()
    if debug:
        rprint(f"[bright_black]Azure Python SDK storage client: {storage_client}[/bright_black]")
    check_assert(storage_client, "Azure Python SDK storage client created")
    from azure.mgmt.storage import StorageManagementClient

    check_assert(isinstance(storage_client, StorageManagementClient), "Azure Python SDK storage client is StorageManagementClient")
    storage_accounts = list(storage_client.storage_accounts.list())
    if debug:
        rprint(f"[bright_black]Azure Python SDK storage accounts: {[account.name for account in storage_accounts]}[/bright_black]")
    check_assert(storage_accounts, "Azure Python SDK storage accounts listed")
    account_idx = [i for i, a in enumerate(storage_accounts) if a.name == account]
    check_assert(len(account_idx) == 1, "Skyplane storage account exists")
    account_details = storage_accounts[account_idx[0]]
    if debug:
        rprint(f"[bright_black]Skyplane storage account: {account_details}[/bright_black]")
    account_subscription = account_details.id.split("/")[2]
    if debug:
        rprint(f"[bright_black]Skyplane storage account subscription: {account_subscription}[/bright_black]")

    # check UMI has access to storage account via Python SDK
    rprint(f"\n{hline}\n[bold]Checking Azure storage account access via UMI...[/bold]\n{hline}")
    # list UMI roles
    cli = f"az role assignment list --assignee {cloud_config.azure_principal_id} --all"
    retcode, stdout, stderr = run_cmd(cli.split(), debug=debug)
    check_assert(retcode == 0, "Azure CLI UMI roles listed", debug_msg=stderr)
    if debug:
        rprint(f"[bright_black]Azure CLI UMI roles: {stdout}[/bright_black]")
    roles = json.loads(stdout)
    role_idx = [i for i, r in enumerate(roles) if r["scope"] == f"/subscriptions/{account_subscription}"]
    check_assert(len(role_idx) >= 1, "Skyplane storage account role assigned to UMI")
    role_names = [roles[i]["roleDefinitionName"] for i in role_idx]
    rprint(f"[bright_black]Skyplane storage account roles: {role_names}[/bright_black]")
    check_assert("Storage Blob Data Contributor" in role_names, "Skyplane storage account has Blob Data Contributor role assigned to UMI")
    check_assert("Storage Account Contributor" in role_names, "Skyplane storage account has Account Contributor role assigned to UMI")

    # check access to container via Python SDK
    rprint(f"\n{hline}\n[bold]Checking Azure container access[/bold]\n{hline}")
    container_client = auth.get_container_client(account, container)
    if debug:
        rprint(f"[bright_black]Azure Python SDK container client: {container_client}[/bright_black]")
    check_assert(container_client, "Azure Python SDK container client created")
    from azure.storage.blob import ContainerClient

    check_assert(isinstance(container_client, ContainerClient), "Azure Python SDK container client is ContainerClient")

    # check skyplane AzureBlobInterface
    rprint(f"\n{hline}\n[bold]Checking Skyplane AzureBlobInterface[/bold]\n{hline}")
    from skyplane.obj_store.azure_blob_interface import AzureBlobInterface

    iface = AzureBlobInterface(account, container)
    print(iface.container_client.get_container_properties())

    # write object temorarily
    # check skyplane AzureBlobInterface
    print(f"\n{hline}\n[bold]Checking Skyplane AzureBlobInterface[/bold]\n{hline}")
    from skyplane.obj_store.azure_blob_interface import AzureBlobInterface

    iface = AzureBlobInterface(account, container)
    print(iface.container_client.get_container_properties())

    # check if writeable
    rprint(f"\n{hline}\n[bold]Checking Skyplane AzureBlobInterface write access[/bold]\n{hline}")
    import tempfile
    import random
    import string

    def generate_random_string(length):
        """Generate a random string of given length"""
        letters = string.ascii_letters
        return "".join(random.choice(letters) for _ in range(length))

    def create_temp_file(size):
        """Create a temporary file with random data"""
        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            temp_file_path = temp_file.name
            random_data = generate_random_string(size)
            temp_file.write(random_data.encode())
        return temp_file_path

    tmp_file = create_temp_file(1024)
    tmp_object = f"skyplane-{uuid.uuid4()}"
    iface.upload_object(tmp_file, tmp_object)
    iface.delete_objects([tmp_object])


@app.command()
def gcp_check(
    bucket: str = typer.Argument(..., help="GCP bucket to check access for"),
    debug: bool = typer.Option(False, "--debug", help="Print debug info"),
):
    hline = "=" * 80
    rprint(f"{hline}\n[bold]Checking Skyplane configuration...[/bold]\n{hline}")
    if debug:
        rprint(f"[bright_black]Skyplane config: {cloud_config}[/bright_black]")
    check_assert(cloud_config.gcp_project_id, "GCP project ID set")
    check_assert(cloud_config.gcp_enabled, "GCP enabled")
    if debug:
        rprint(f"[bright_black]GCP project ID: {cloud_config.gcp_project_id}[/bright_black]")

    # check that GCloud CLI works
    rprint(f"\n{hline}\n[bold]Checking GCloud CLI...[/bold]\n{hline}")
    cli = "gcloud auth list"
    retcode, stdout, stderr = run_cmd(cli.split(), debug=debug)
    check_assert(retcode == 0, "GCloud CLI authenticated", debug_msg=stderr)
    if debug:
        rprint(f"[bright_black]GCloud CLI auth list: {stdout}[/bright_black]")
    check_assert("ACTIVE" in stdout, "GCloud CLI authenticated")

    # check that GCP Python SDK works
    rprint(f"\n{hline}\n[bold]Checking GCP Python SDK...[/bold]\n{hline}")
    auth = compute.GCPAuthentication(cloud_config)
    if debug:
        rprint(f"[bright_black]GCP Python SDK auth: {auth}[/bright_black]")
    check_assert(auth, "GCP Python SDK auth created")
    cred = auth.credentials
    sa_cred = auth.service_account_credentials
    if debug:
        rprint(f"[bright_black]GCP Python SDK credentials: {cred}[/bright_black]")
        rprint(f"[bright_black]GCP Python SDK service account credentials: {sa_cred}[/bright_black]")
    check_assert(cred, "GCP Python SDK credentials created")
    check_assert(sa_cred, "GCP Python SDK service account credentials created")


@app.command()
def azure_get_valid_skus(
    regions: List[str] = typer.Option(compute.AzureCloudProvider.region_list(), "--regions", "-r"),
    prefix: str = typer.Option("", "--prefix", help="Filter by prefix"),
    top_k: int = typer.Option(-1, "--top-k", help="Print top k entries"),
):
    auth = compute.AzureAuthentication()
    client = auth.get_compute_client()

    def get_skus(region):
        valid_skus = []
        for sku in client.resource_skus.list(filter="location eq '{}'".format(region)):
            if len(sku.restrictions) == 0 and (not prefix or sku.name.startswith(prefix)):
                valid_skus.append(sku.name)
        return set(valid_skus)

    result = do_parallel(get_skus, regions, spinner=True, desc="Query SKUs")

    sku_regions = defaultdict(set)
    for region, skus in result:
        for sku in skus:
            sku_regions[sku].add(region)

    # print top-k entries (if not -1)
    sorted_top_keys = sorted(sku_regions.keys(), key=lambda x: len(sku_regions[x]), reverse=True)
    if top_k > 0:
        sorted_top_keys = sorted_top_keys[:top_k]
    for sku in sorted_top_keys:
        typer.secho(f"{sku} in {len(sku_regions[sku])} regions: {list(sorted(sku_regions[sku]))}")



from importlib.resources import path

from skyplane.utils import logger

try:
    import pandas as pd
except ImportError:
    pd = None
    logger.warning("pandas not installed, will not be able to load transfer costs")


class AWSPricing:
    def __init__(self):
        self._transfer_df = None

    @property
    def transfer_df(self):
        if pd:
            if self._transfer_df is None:
                with path("skyplane.data", "aws_transfer_costs.csv") as transfer_cost_path:
                    self._transfer_df = pd.read_csv(transfer_cost_path).set_index(["src", "dst"])

            return self._transfer_df
        else:
            return None

    def get_transfer_cost(self, src_key, dst_key, premium_tier=True):
        assert premium_tier, "AWS transfer cost is only available for premium tier"
        transfer_df = self.transfer_df
        if transfer_df is None:
            return None
        else:
            src_provider, src = src_key.split(":")
            dst_provider, dst = dst_key.split(":")

            assert src_provider == "aws"
            if dst_provider == "aws":
                if (src, dst) in transfer_df.index:
                    return transfer_df.loc[src, dst]["cost"]
                else:
                    logger.warning(f"No transfer cost found for {src_key} -> {dst_key}, using max of {src}")
                    src_rows = transfer_df.loc[src]
                    src_rows = src_rows[src_rows.index != "internet"]
                    return src_rows.max()["cost"]
            else:
                return transfer_df.loc[src, "internet"]["cost"]



"""
Created on 21 Feb 2023

@author: gilvernik
"""

import os


def load_yaml_config(config_filename):
    import yaml

    try:
        with open(config_filename, "r") as config_file:
            data = yaml.safe_load(config_file)
    except FileNotFoundError:
        data = {}

    return data


def dump_yaml_config(config_filename, data):
    import yaml

    if not os.path.exists(os.path.dirname(config_filename)):
        os.makedirs(os.path.dirname(config_filename))

    with open(config_filename, "w") as config_file:
        yaml.dump(data, config_file, default_flow_style=False)


def delete_yaml_config(config_filename):
    if os.path.exists(config_filename):
        os.remove(config_filename)



from multiprocessing import Queue


class GatewayQueue:
    def __init__(self, maxsize=10000):
        self.q = Queue(maxsize)
        self.handles = []

    def register_handle(self, requester_handle):
        self.handles.append(requester_handle)

    def put(self, chunk_req):
        self.q.put(chunk_req)

    def put_nowait(self, chunk_req):
        self.q.put_nowait(chunk_req)

    def pop(self, requester_handle=None):
        self.q.get()

    def get_nowait(self, requester_handle=None):
        return self.q.get_nowait()

    def get_handles(self):
        return self.handles

    def size(self):
        return self.q.qsize()


class GatewayANDQueue(GatewayQueue):
    def __init__(self, maxsize=10000):
        self.q = {}
        self.maxsize = maxsize
        self.temp_q = Queue(maxsize)  # temporarily store value

    def register_handle(self, requester_handle):
        # create a queue per handle (operator)
        self.q[requester_handle] = GatewayQueue(self.maxsize)

    def get_handles(self):
        return list(self.q.keys())

    def get_handle_queue(self, requester_handle):
        return self.q[requester_handle]

    def put(self, chunk_req):
        # place chunk in all downstream operators queues
        for handle in self.q.keys():
            print("add queue", handle, self.q[handle].size())
            # self.q[handle].put(copy.deepcopy(chunk_req))
            self.q[handle].put(chunk_req)

    def put_nowait(self, chunk_req):
        raise ValueError("GatewayANDQueue cannot be the first queue in a pipeline")

    def pop(self, requester_handle):
        self.q[requester_handle].get()

    def get_nowait(self, requester_handle):
        return self.q[requester_handle].get_nowait()



from typing import List, Optional
from collections import defaultdict
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn, DownloadColumn, TransferSpeedColumn, TimeRemainingColumn
from skyplane import exceptions
from skyplane.chunk import Chunk
from skyplane.cli.impl.common import console
from skyplane.utils.definitions import format_bytes
from skyplane.api.tracker import TransferHook


class ProgressBarTransferHook(TransferHook):
    """Transfer hook for multi-destination transfers."""

    def on_dispatch_start(self):
        return

    def __init__(self, dest_region_tags: List[str]):
        self.spinner = Progress(
            SpinnerColumn(),
            TextColumn("Dispatching chunks...{task.description}"),
            BarColumn(),
            DownloadColumn(binary_units=True),
            transient=True,
        )
        self.dest_region_tags = dest_region_tags
        self.pbar = None  # map between region_tag and progress bar
        self.transfer_task = {}
        self.chunks_dispatched = 0
        self.bytes_dispatched = 0
        self.chunks_completed = defaultdict(int)
        self.bytes_completed = defaultdict(int)
        self.dispatch_task = self.spinner.add_task("", total=None)
        self.spinner.start()

    def on_chunk_dispatched(self, chunks: List[Chunk]):
        # update bytes_dispatched
        if len(chunks) == 0:
            self.bytes_dispatched = 0
        else:
            self.bytes_dispatched += sum([chunk.chunk_length_bytes for chunk in chunks])
            self.chunks_dispatched += len(chunks)
        # rerender spinners with updated text "Dispatching chunks (~{format_bytes(self.bytes_dispatched)} dispatched)"
        self.spinner.update(
            self.dispatch_task, description=f" {self.chunks_dispatched} chunks (~{format_bytes(self.bytes_dispatched)} dispatched)"
        )

    def on_dispatch_end(self):
        self.spinner.stop()
        self.pbar = Progress(
            SpinnerColumn(),
            TextColumn("Transfer progress{task.description}"),
            BarColumn(),
            DownloadColumn(binary_units=True),
            TransferSpeedColumn(),
            TimeRemainingColumn(),
            transient=True,
        )
        for region_tag in self.dest_region_tags:
            self.transfer_task[region_tag] = self.pbar.add_task(region_tag, total=self.bytes_dispatched)
        self.pbar.start()

    def on_chunk_completed(self, chunks: List[Chunk], region_tag: Optional[str] = None):
        assert region_tag is not None, f"Must specify region tag for progress bar"
        self.chunks_completed[region_tag] += len(chunks)
        self.bytes_completed[region_tag] += sum([chunk.chunk_length_bytes for chunk in chunks])
        self.pbar.update(self.transfer_task[region_tag], completed=self.bytes_completed[region_tag])

    def on_transfer_end(self):
        self.pbar.stop()

    def on_transfer_error(self, error):
        console.log(error)
        raise exceptions.SkyplaneGatewayException("Transfer failed with error", error)



import os
from pathlib import Path

from skyplane.gateway_version import gateway_version

KB = 1024
MB = 1024 * 1024
GB = 1024 * 1024 * 1024


def format_bytes(bytes_int: int):
    if bytes_int < KB:
        return f"{bytes_int}B"
    elif bytes_int < MB:
        return f"{bytes_int / KB:.2f}KB"
    elif bytes_int < GB:
        return f"{bytes_int / MB:.2f}MB"
    else:
        return f"{bytes_int / GB:.2f}GB"


is_gateway_env = os.environ.get("SKYPLANE_IS_GATEWAY", None) == "1"


def gateway_docker_image():
    return "public.ecr.aws/s6m1p0n8/skyplane:" + gateway_version


tmp_log_dir = Path("/tmp/skyplane")
